"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020-2025 Lukas Hörtnagl

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""

import sys
from pathlib import Path

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from diive.pkgs.outlierdetection.zscore import zScoreRolling

from dyco import loop, plot


class AnalyzeLags:
    """
    Analyze lag search results and create look-up table for lag-time normalization

    * Creates LUT for daily default time lags (aggregated) needed in Phase 1 and Phase 2.
    * Creates LUT for instantaneous time lags needed in Phase 3.

    """

    def __init__(self,
                 lgs_num_iter,
                 outdirs,
                 target_lag,
                 logger,
                 outlier_winsize,
                 outlier_thres_zscore,
                 lags: pd.DataFrame):

        self.lgs_num_iter = lgs_num_iter
        self.outdirs = outdirs
        self.target_lag = target_lag
        self.logger = logger
        self.lags = lags.copy()

        self.outlier_winsize = outlier_winsize
        self.outlier_thres_zscore = outlier_thres_zscore

        self.lut_lag_times_df = pd.DataFrame()
        self.lut_available = False

        self.run()

    def get_lut(self):
        return self.lut_lag_times_df

    def run(self):
        """Run the lag analysis"""

        # segment_lagtimes_last_iteration_df['PEAK-COVABSMAX_SHIFT']

        # Make lookup table from aggregated daily lags
        self.lut_lag_times_df, self.lut_available = self.make_lut_agg()

        if self.lut_available:
            self.logger.info(f"Finished creating look-up table for default lag times and normalization correction")
        else:
            self.logger.critical(f"(!) Look-up Table for default lag times and normalization correction is empty, "
                                 f"stopping script.")
            sys.exit()

        if self.outdirs:
            self.save_lut(lut=self.lut_lag_times_df,
                          outdir=self.outdirs['7_time_lags_lookup_table'],
                          outfile='LUT_default_agg_time_lags')
            self.plot_segment_lagtimes_with_agg_default()

    @staticmethod
    def plot_final_instantaneous_lagtimes(outdir, phase, df):
        """
        Read and plot final lag search result: the instantaneous time lags

        Parameters
        ----------
        outdir: Path
            Output folder where results are stored
        phase: int
            Phase of the processing chain.
        df: pandas DataFrame

        Returns
        -------
        None
        """

        # Get data
        lagsearch_start = int(df['LAGSEARCH_START'].unique()[0])
        lagsearch_end = int(df['LAGSEARCH_END'].unique()[0])
        abs_limit = int(df['ABS_LIMIT'].unique()[0])

        # Plot
        gs, fig, ax = plot.setup_fig_ax()

        # Accepted reference lags
        ax.plot_date(pd.to_datetime(df.index), df['INSTANTANEOUS_LAG'],
                     alpha=1, fmt='o', ms=6, color='black', lw=0, ls='-',
                     label=f'final reference time lag (absolute limit {abs_limit})', markeredgecolor='None', zorder=100)

        # Found lags in Phase 3
        ax.plot_date(pd.to_datetime(df.index), df['PEAK-COVABSMAX_SHIFT'],
                     alpha=1, fmt='o', ms=12, color='#FFC107', lw=0, ls='-', markeredgecolor='None', zorder=99,
                     label=f'found Phase 3 time lag (search between {lagsearch_start} and {lagsearch_end})')

        # Marks lags that were outside limit and therefore set to the default lag
        set_to_default_lags = df.loc[df['SET_TO_DEFAULT'] == True, ['INSTANTANEOUS_LAG']]
        ax.plot_date(pd.to_datetime(set_to_default_lags.index), set_to_default_lags['INSTANTANEOUS_LAG'],
                     alpha=1, fmt='o', ms=12, color='#8BC34A', lw=0, ls='-', markeredgecolor='None', zorder=99,
                     label=f'found Phase 3 time lag was set to default')

        plot.default_format(ax=ax, label_color='black', fontsize=12,
                            txt_xlabel='segment date', txt_ylabel='lag', txt_ylabel_units='[records]')

        txt_info = f"PHASE {phase}: FINAL REFERENCE TIME LAGS"
        font = {'family': 'sans-serif', 'color': 'black', 'weight': 'bold', 'size': 20, 'alpha': 1, }
        ax.set_title(txt_info, fontdict=font)

        ax.axhline(0, color='black', ls='-', lw=1, label='default lag', zorder=98)
        limit = df['ABS_LIMIT'].unique()[0]
        ax.axhline(limit, color='#d32f2f', ls='--', lw=1, label='upper lag acceptance limit', zorder=98)
        ax.axhline(limit * -1, color='#7B1FA2', ls='--', lw=1, label='lower lag acceptance limit', zorder=98)
        font = {'family': 'sans-serif', 'size': 10}
        ax.legend(frameon=True, loc='upper right', prop=font).set_zorder(100)

        # Automatic tick locations and formats
        locator = mdates.AutoDateLocator(minticks=5, maxticks=20)
        formatter = mdates.ConciseDateFormatter(locator)
        ax.xaxis.set_major_locator(locator)
        ax.xaxis.set_major_formatter(formatter)

        # Save
        outpath = outdir / f'PHASE-{phase}_FINAL_TIME_LAGS_FOR_REFERENCE_VAR'
        # print(f"Saving time series of found segment lag times in {outpath} ...")
        fig.savefig(f"{outpath}.png", format='png', bbox_inches='tight', facecolor='w',
                    transparent=True, dpi=150)
        plt.close(fig)

    def plot_segment_lagtimes_with_agg_default(self):
        """
        Read and plot found time lags from all or only very last iteration(s),
        used in Phase 1 and Phase 2.
        """
        loop.Loop.plot_segment_lagtimes_ts(segment_lagtimes_df=self.lags,
                                           outdir=self.outdirs[f'7_time_lags_lookup_table'],
                                           show_all=False,
                                           overlay_default=True,
                                           overlay_default_df=self.lut_lag_times_df,
                                           overlay_target_val=self.target_lag)

    def save_lut(self, lut, outdir, outfile):
        """
        Save the look-up table for lag correction to csv file

        Parameters
        ----------
        lut: pandas DataFrame
            Contains look-up table information.
        outdir: Path
            Output folder for outfile.
        outfile: str
            Filename without extension.

        Returns
        -------
        None
        """
        outpath = outdir / outfile
        lut.to_csv(f"{outpath}.csv")

    def make_lut_instantaneous(self, segment_lagtimes_df: pd.DataFrame, default_lag: int):
        """
        Generate instantaneous lag look-up table that contains the found lag time
        for each averaging interval

        Used in Phase 3.

        Parameters
        ----------
        segment_lagtimes_df: pandas DataFrame
            Contains found lag times for each segment.
        default_lag: int
            The median of high-quality peaks is moved to target_lag.

        Returns
        -------
        pandas DataFrame with default lag times for each day

        """
        # Prepare df
        _segment_lagtimes_df = segment_lagtimes_df.set_index('file_date')
        lut_df = _segment_lagtimes_df[['PEAK-COVABSMAX_SHIFT']].copy()  # file_date needed for look-up
        # lut_df = segment_lagtimes_df[['file_date', 'PEAK-COVABSMAX_SHIFT']].copy()  # file_date needed for look-up
        # lut_df.set_index('file_date', inplace=True)
        lut_df['LAGSEARCH_START'] = np.nan
        lut_df['LAGSEARCH_END'] = np.nan
        lut_df['ABS_LIMIT'] = np.nan
        lut_df['DEFAULT_LAG'] = np.nan
        lut_df['SET_TO_DEFAULT'] = np.nan
        lut_df['INSTANTANEOUS_LAG'] = np.nan
        lut_df['DEFAULT_LAG'] = default_lag

        lut_df['ABS_LIMIT'] = 50
        lut_df['LAGSEARCH_START'] = _segment_lagtimes_df['lagsearch_start']
        lut_df['LAGSEARCH_END'] = _segment_lagtimes_df['lagsearch_end']

        # Replace found lags above absolute threshold with default lag, keep others
        filter_set_to_default = (lut_df['PEAK-COVABSMAX_SHIFT'].abs() > lut_df['ABS_LIMIT']) | \
                                (lut_df['PEAK-COVABSMAX_SHIFT'].isnull())
        lut_df['SET_TO_DEFAULT'] = filter_set_to_default
        lut_df.loc[filter_set_to_default, 'INSTANTANEOUS_LAG'] = lut_df.loc[filter_set_to_default, 'DEFAULT_LAG']
        lut_df.loc[~filter_set_to_default, 'INSTANTANEOUS_LAG'] = lut_df.loc[
            ~filter_set_to_default, 'PEAK-COVABSMAX_SHIFT']

        self.logger.info(f"Created look-up table for {len(lut_df.index)} dates")
        self.logger.info(f"    First date: {lut_df.index[0]}    Last date: {lut_df.index[-1]}")

        # Check for gaps
        missing_df = self.check_missing(df=lut_df,
                                        col='INSTANTANEOUS_LAG')
        if missing_df.empty:
            # All lags available
            pass
        else:
            # Some or all lags missing, fill with default lag
            self.logger.warning(f"No lag was available for dates: {missing_df.index.to_list()}")
            self.logger.warning(f"Filling missing lags with default lag, affected dates: {missing_df.index.to_list()}")
            lut_df['INSTANTANEOUS_LAG'].fillna(lut_df['DEFAULT_LAG'])

        lut_available = True
        return lut_df, lut_available

    def _remove_outliers(self, peaks_hq_S):
        self.outlier_winsize = int(len(peaks_hq_S) / 70) if not self.outlier_winsize else self.outlier_winsize
        zsr = zScoreRolling(
            series=peaks_hq_S,
            thres_zscore=self.outlier_thres_zscore,
            winsize=self.outlier_winsize,
            showplot=True,
            plottitle="z-score in a rolling window",
            verbose=True)
        zsr.calc(repeat=True)

        fig = zsr.fig

        # Save
        outdir = self.outdirs[f'7_time_lags_lookup_table']
        outfile = f"TIMESERIES-PLOT_segment_lag_times_FINAL_outlierRemoved"
        outpath = outdir / outfile
        # print(f"Saving time series of found segment lag times in {outpath} ...")
        fig.savefig(f"{outpath}.png", format='png', bbox_inches='tight', facecolor='w',
                    transparent=True, dpi=150)
        plt.close(fig)

        flag = zsr.get_flag()
        peaks_hq_S_cleaned = peaks_hq_S.loc[flag == 0].copy()
        return peaks_hq_S_cleaned

    def make_lut_agg(self):
        """
        Generate aggregated look-up table that contains the default lag time for each day

        Default lag times are determined by
            (1) pooling data of the current day with data of the day before and
                the day after,
            (2) calculating the median of the pooled data.

        Used in Phase 1 and Phase 2.

        Parameters
        ----------
        segment_lagtimes_df: pandas DataFrame
            Contains found lag times for each segment.
        target_lag: int
            The median of high-quality peaks is moved to target_lag.

        Returns
        -------
        pandas DataFrame with default lag times for each day

        """

        # Initiate empty LUT
        lut_df = pd.DataFrame()

        # Find high-quality covariance peaks
        peaks_hq_S = self.get_hq_peaks()
        peaks_hq_S = peaks_hq_S.sort_index(inplace=False)
        peaks_hq_S_cleaned = self._remove_outliers(peaks_hq_S)

        if peaks_hq_S_cleaned.empty:
            lut_available = False
            return lut_df, lut_available

        unique_dates = np.unique(peaks_hq_S_cleaned.index.date)
        for this_date in unique_dates:
            from_date = this_date - pd.Timedelta('2D')
            to_date = this_date + pd.Timedelta('2D')
            filter_around_this_day = (peaks_hq_S_cleaned.index.date > from_date) & \
                                     (peaks_hq_S_cleaned.index.date <= to_date)
            subset = peaks_hq_S_cleaned[filter_around_this_day]
            num_vals = len(subset)

            if num_vals >= 10:
                # print(f"{this_date}    {num_vals}    {subset.median()}")
                lut_df.loc[this_date, 'median'] = subset.median()
            else:
                lut_df.loc[this_date, 'median'] = np.nan

            lut_df.loc[this_date, 'counts'] = subset.count()
            lut_df.loc[this_date, 'from'] = from_date
            lut_df.loc[this_date, 'to'] = to_date

        # Detect first and last dates for look-up table
        firstdate = pd.to_datetime(self.lags['start'].iloc[0]).date()
        lastdate = pd.to_datetime(self.lags['end'].iloc[-1]).date()

        # Make sure LUT has all dates between start and end dates
        fullrange = pd.date_range(firstdate, lastdate, freq='d')
        lut_df = lut_df.reindex(fullrange)

        # Filling missing median values with rolling median in a 5-day window, centered
        n_missing_medians = lut_df['median'].isnull().sum()
        if n_missing_medians > 0:
            lut_df['median'] = lut_df['median'].fillna(
                lut_df['median'].rolling(window=5, min_periods=1, center=True).median())
            self.logger.info(f"(!)WARNING: Missing look-up values for {n_missing_medians} days "
                             f"were gap-filled with the median value of the 2 preceding "
                             f"and the 2 following days.")

        lut_df['target_lag'] = self.target_lag
        lut_df['correction'] = -1 * (lut_df['target_lag'] - lut_df['median'])

        self.logger.info(f"Created look-up table for {len(lut_df.index)} dates")
        self.logger.info(f"    First date: {lut_df.index[0]}    Last date: {lut_df.index[-1]}")

        # Fill small gaps for correction
        missing_df = self.check_missing(df=lut_df, col='correction')
        if not missing_df.empty:
            lut_df['correction'] = lut_df['correction'].ffill(limit=1)
            lut_df['correction'] = lut_df['correction'].bfill(limit=1)
            self.logger.warning(f"(!) Missing corrections for days: {missing_df.index.to_list()}\n"
                                f"(!) Using correction values from directly adjacent day.")

        # Fill gaps in 'correction'
        missing_df = self.check_missing(df=lut_df, col='correction')
        if not missing_df.empty:
            self.logger.warning(
                f"(!) No correction could be generated from data for dates: {missing_df.index.to_list()}")

        # lut_df['correction'] = lut_df['correction'].fillna(method='ffill', inplace=False, limit=1)  # deprecated

        lut_available = True
        return lut_df, lut_available

    def check_missing(self, df, col):
        """
        Check for missing values in data rows

        Parameters
        ----------
        df: pandas DataFrame
        col: str
            Column name of the variable that is checked for missing data.

        Returns
        -------
        pandas DataFrame that only contains data rows where a value for col is missing
        """
        filter_missing = df[col].isnull()
        missing_df = df[filter_missing]
        return missing_df

    def get_hq_peaks(self):
        """
        Detect high-quality covariance peaks in results from last lag search iteration

        High-quality means that during the covariance calculations the max covariance
        peak and the automatically detected peak yielded the same results, i.e. the
        same record.

        Parameters
        ----------
        df: pandas DataFrame containing results from the last lag search iteration

        Returns
        -------
        pandas Series of high-quality lag times, given as number of records

        """
        df = self.lags.copy()
        df.set_index('start', inplace=True)
        df.index = pd.to_datetime(df.index, format="mixed")
        peaks_hq_S = df.loc[df['PEAK-COVABSMAX_SHIFT'] == df['PEAK-AUTO_SHIFT'], 'PEAK-COVABSMAX_SHIFT']
        peaks_hq_S.index = peaks_hq_S.index.to_pydatetime()  # Convert to DatetimeIndex
        return peaks_hq_S



"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020-2025 Lukas Hörtnagl

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""

import datetime as dt
from pathlib import Path

import numpy as np
import pandas as pd
from diive.core.io.files import load_parquet


def read_segment_lagtimes_file(filepath):
    """
    Read file.

    Reading segment covariances and lag search results for each segment.
    Can be used for all text files for which the .read_csv args are valid.

    Parameters
    ----------
    filepath: str

    Returns
    -------
    pandas DataFrame

    """
    # parse = lambda x: dt.datetime.strptime(x, '%Y%m%d%H%M%S')
    found_lags_df = pd.read_csv(filepath,
                                skiprows=None,
                                header=0,
                                # names=header_cols_list,
                                # na_values=-9999,
                                encoding='utf-8',
                                delimiter=',',
                                # mangle_dupe_cols=True,
                                # keep_date_col=False,
                                parse_dates=False,
                                # date_parser=parse,
                                index_col=0,
                                dtype=None,
                                engine='c')
    return found_lags_df


def read_raw_data(filepath, data_timestamp_format):
    """
    Read raw data files

    Parameters
    ----------
    filepath: Path
        Full path to raw data file.
    data_timestamp_format: str
        Datetime format of the timestamp in each data row in the raw data file.

    Returns
    -------
    pandas DataFrame that contains raw data from the file in filepath
    """

    file_ext = Path(filepath).suffix

    if file_ext == '.csv':
        data_df = read_raw_data_csv(filepath, data_timestamp_format)

    elif file_ext == '.parquet':
        data_df = load_parquet(filepath, output_middle_timestamp=False, sanitize_timestamp=False)

    else:
        raise Exception('File extension must be ".csv" or ".parquet"')

    return data_df


def read_raw_data_csv(filepath, data_timestamp_format):
    header_rows_list = [0]
    skip_rows_list = []
    header_section_rows = [0]

    num_data_cols = \
        length_data_cols(filepath=filepath,
                         header_rows_list=header_rows_list,
                         skip_rows_list=skip_rows_list)

    num_header_cols, header_cols_df = \
        length_header_cols(filepath=filepath,
                           header_rows_list=header_rows_list,
                           skip_rows_list=skip_rows_list)

    more_data_cols_than_header_cols, num_missing_header_cols = \
        data_vs_header(num_data_cols=num_data_cols,
                       num_header_cols=num_header_cols)

    header_cols_list = \
        generate_missing_cols(header_cols_df=header_cols_df,
                              more_data_cols_than_header_cols=more_data_cols_than_header_cols,
                              num_missing_header_cols=num_missing_header_cols)

    if data_timestamp_format:
        # parse = lambda x: dt.datetime.strptime(x, data_timestamp_format)  # now deprecated
        # date_parser = parse  # now deprecated
        parse_dates = True
        index_col = 0
    else:
        # date_parser = None  # now deprecated
        parse_dates = False
        index_col = None

    data_df = pd.read_csv(filepath,
                          skiprows=header_section_rows,
                          header=None,
                          names=header_cols_list,
                          na_values=-9999,
                          encoding='utf-8',
                          delimiter=',',
                          # mangle_dupe_cols=True,  # now deprecated
                          # keep_date_col=False,  # now deprecated
                          parse_dates=parse_dates,
                          # date_parser=date_parser,  # now deprecated
                          date_format=data_timestamp_format,
                          index_col=index_col,
                          dtype=None,
                          engine='c',
                          nrows=None)

    return data_df


def add_data_stats(df, true_resolution, filename, files_overview_df, found_records, fnm_date_format):
    """
    Collect additional info about raw data file

    Parameters
    ----------
    df: pandas DataFrame
        Raw data from the file.
    true_resolution: float
        Time resolution of the raw data records in seconds, e.g. 0.05 for 20 Hz data.
    filename: str
        Filename of the raw data file, without extension.
    files_overview_df: pandas DataFrame
        Overview of all raw data files, with stats.
    found_records: int
        Number of records in the raw data file.
    fnm_date_format: str
        Datetime format of the datetime info in the raw data filename.

    Returns
    -------
    pandas DataFrame with additional info for current raw data file
    """
    # Detect overall frequency
    data_duration = found_records * true_resolution
    data_freq = np.float64(found_records / data_duration)

    idx = dt.datetime.strptime(filename, fnm_date_format)  # Use filename datetime info as index

    files_overview_df.loc[idx, 'first_record'] = df.index[0]
    files_overview_df.loc[idx, 'last_record'] = df.index[-1]
    files_overview_df.loc[idx, 'file_duration'] = (df.index[-1] - df.index[0]).total_seconds()
    files_overview_df.loc[idx, 'found_records'] = found_records
    files_overview_df.loc[idx, 'data_freq'] = data_freq

    return files_overview_df


def generate_missing_cols(header_cols_df, more_data_cols_than_header_cols, num_missing_header_cols):
    """
    Insert additional column names in data header

    Additional columns are created if the number of data columns
    does not match the number of header columns.

    Parameters
    ----------
    header_cols_df: pandas DataFrame
        A small DataFrame that only contains the header columns.
    more_data_cols_than_header_cols: bool
        True if more data columns than header columns were found in
        the data file. This can happen due to irregularities during
        raw data collection.
    num_missing_header_cols: int
        Number of missing header columns in comparison to data columns.

    Returns
    -------
    list of header columns that contains labels for additionally created columns
    """
    # Generate missing header columns if necessary
    header_cols_list = header_cols_df.columns.to_list()
    generated_missing_header_cols_list = []
    if more_data_cols_than_header_cols:
        for m in list(range(1, num_missing_header_cols + 1)):
            missing_col = (f'unknown_{m}')
            generated_missing_header_cols_list.append(missing_col)
            header_cols_list.append(missing_col)
    return header_cols_list


def length_data_cols(filepath, header_rows_list, skip_rows_list):
    """
    Check number of columns of the first data row after the header part

    Parameters
    ----------
    filepath: Path
        Path to raw data file
    header_rows_list: list
        List of integers that give the row positions of the header lines
    skip_rows_list: list
        List of skipped rows

    Returns
    -------
    Number of data columns
    """
    skip_num_lines = len(header_rows_list) + len(skip_rows_list)
    first_data_row_df = pd.read_csv(filepath,
                                    skiprows=skip_num_lines,
                                    header=None,
                                    nrows=1)

    return first_data_row_df.columns.size


def length_header_cols(filepath, header_rows_list, skip_rows_list):
    """
    Check number of columns of the header part

    Parameters
    ----------
    filepath: Path
        Path to raw data file
    header_rows_list: list
        List of integers that give the row positions of the header lines
    skip_rows_list: list
        List of skipped rows

    Returns
    -------
    Number of header columns and a pandas DataFrame that contains only the header columns
    """
    header_cols_df = pd.read_csv(filepath,
                                 skiprows=skip_rows_list,
                                 header=header_rows_list,
                                 nrows=0)
    return header_cols_df.columns.size, header_cols_df


def data_vs_header(num_data_cols, num_header_cols):
    """
    Check if there are more data columns than header columns

    Parameters
    ----------
    num_data_cols: int
        Number of data columns
    num_header_cols: int
        Number of header columns

    Returns
    -------
    more_data_cols_than_header_cols: bool
        True if number of data columns > number of header columns
    num_missing_header_cols: int
        Number of missing header columns compared to number of data columns
    """
    if num_data_cols > num_header_cols:
        more_data_cols_than_header_cols = True
        num_missing_header_cols = num_data_cols - num_header_cols
    else:
        more_data_cols_than_header_cols = False
        num_missing_header_cols = 0
    return more_data_cols_than_header_cols, num_missing_header_cols



"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020-2025 Lukas Hörtnagl

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.signal import find_peaks

from dyco import plot


class AdjustLagsearchWindow:
    """

    Adjust the time window for lag search that is used in the next iteration

    """

    def __init__(self, series, iteration, plot=True, hist_num_bins=30, remove_fringe_bins=True,
                 perc_threshold=0.9, outdir=None):
        self.series = series.dropna()  # NaNs yield error in histogram
        self.numvals_series = self.series.size
        self.outdir = outdir
        self.iteration = iteration
        self.plot = plot
        self.hist_num_bins = hist_num_bins
        self.remove_fringe_bins = remove_fringe_bins
        self.perc_threshold = perc_threshold

        self.run()

    def run(self):
        """Start adjustment of lagsearch time window"""
        self.lgs_winsize_adj, self.peak_max_count_idx, self.start_idx, self.end_idx, \
            self.counts, self.divisions, self.peak_most_prom_idx = \
            self.find_hist_peaks()

        if self.plot:
            self.plot_results_hist(hist_bins=self.divisions)

    def find_hist_peaks(self):
        """Find peak in histogram of found lag times."""

        # Make histogram of found lag times, remove fringe bins at start and end
        counts, divisions = self.calc_hist(series=self.series, bins=self.hist_num_bins,
                                           remove_fringe_bins=self.remove_fringe_bins)

        # Search bin with most found lag times
        peak_max_count_idx = self.search_bin_max_counts(counts=counts)

        # Search most prominent bin
        peak_most_prom_idx = self.search_bin_most_prominent(counts=counts)

        # Adjust lag search time window for next iteration
        lgs_winsize_adj, start_idx, end_idx = self.adjust_lgs_winsize(counts=counts, divisions=divisions,
                                                                      perc_threshold=self.perc_threshold,
                                                                      peak_max_count_idx=peak_max_count_idx,
                                                                      peak_most_prom_idx=peak_most_prom_idx)

        # todo (maybe) Check if most prominent peak is also the max peak
        if peak_most_prom_idx in peak_max_count_idx:
            clear_peak_idx = np.where(peak_max_count_idx == peak_most_prom_idx)
        else:
            clear_peak_idx = False

        lgs_winsize_adj = [divisions[start_idx], divisions[end_idx]]
        # Convert elements in array to integers, needed for indexing
        lgs_winsize_adj = [int(x) for x in lgs_winsize_adj]

        # Expand lag search window size
        # Set minimum size of lag search time window to min. 20 records
        # Zero is also valid, or in case of two positive numbers, the values are both used, +1
        # [+1, +8] --> allowed values are [1, 2, 3, 4, 5, 6, 7, 8] --> range=8-1=7 -> +1=8
        # [-4, +2] --> allowed values are [-4, -3, -2, -1, 0, +1, +2] --> range=2-(-4)=6 -> +1=7
        _range = (lgs_winsize_adj[1] - lgs_winsize_adj[0]) + 1
        while _range < 20:
            # [-8, -3] --> [-9, -2] ... --> finally [-15, 4]
            # [-4, +2] --> [-5, +3] ...
            # [+2, +7] --> [+1, +8] ...
            lgs_winsize_adj[0] = lgs_winsize_adj[0] - 1
            lgs_winsize_adj[1] = lgs_winsize_adj[1] + 1
            _range = (lgs_winsize_adj[1] - lgs_winsize_adj[0]) + 1

        return lgs_winsize_adj, peak_max_count_idx, start_idx, end_idx, counts, divisions, peak_most_prom_idx

    def get(self):
        """
        Adjusted time window for lag search

        Returns
        -------
        lgs_winsize_adj: list
            Contains two elements: start index and end index for lag search, i.e. the
            "search from/to" range for the lag search

        """
        return self.lgs_winsize_adj

    def search_bin_max_counts(self, counts):
        """
        Search histogram peak of maximum counts

        Parameters
        ----------
        counts: array
            Counts in histogram bins

        Returns
        -------
        Index of the histogram peak with maximum counts
        """
        max_count = np.amax(counts)
        peak_max_count_idx = np.where(counts == np.amax(max_count))  # Returns array in tuple
        if len(peak_max_count_idx) == 1:
            peak_max_count_idx = peak_max_count_idx[0]  # Yields array of index or multiple indices
        return peak_max_count_idx

    @staticmethod
    def calc_hist(series=False, bins=20, remove_fringe_bins=False):
        """
        Calculate histogram of found lag times

        Done after each iteration.

        Parameters
        ----------
        series: pandas Series
            Found absolute covariance peaks for each segment
        bins: int or range
            Number of bins for the histogram, can also be given as range
        remove_fringe_bins: bool
            If True, the fringe bins in the histogram will be removed. Important for
            constraining the lag during lag search.

        Returns
        -------
        counts: counts per histogram division
        divisions: histogram divisions
        """
        counts, divisions = np.histogram(series, bins=bins)
        # Remove first and last bins from histogram. In case of unclear lag times
        # data tend to accumulate in these edge regions of the search window.
        if remove_fringe_bins and len(counts) >= 5:
            counts = counts[1:-1]
            divisions = divisions[1:-1]  # Contains start values of bins
        return counts, divisions

    @staticmethod
    def search_bin_most_prominent(counts):
        """
        Search most prominent histogram peak

        Increase prominence until only one single peak is found
        kudos: https://www.kaggle.com/simongrest/finding-peaks-in-the-histograms-of-the-variables

        Parameters
        ----------
        counts: array
            Counts per histogram bin/division

        Returns
        -------
        Index of the most prominent peak
        """
        peak_most_prom_idx = []
        prom = 0  # Prominence for peak finding
        while (len(peak_most_prom_idx) == 0) or (len(peak_most_prom_idx) > 1):
            prom += 1
            if prom > 40:
                peak_most_prom_idx = False
                break
            peak_most_prom_idx, props = find_peaks(counts, prominence=prom)
            # print(f"Prominence: {prom}    Peaks at: {peak_most_prom_idx}")
        if peak_most_prom_idx:
            peak_most_prom_idx = int(peak_most_prom_idx)
        return peak_most_prom_idx

    def adjust_lgs_winsize(self, counts, divisions, perc_threshold, peak_max_count_idx, peak_most_prom_idx):
        """
        Set new time window for next lag search, based on previous results

        Includes more and more bins around the bin where most lag times were found
        until a threshold is reached.

        Parameters
        ----------
        counts: numpy array
            Counts per histogram division
        divisions: numpy array
            Histogram divisions
        perc_threshold: float
            Cumulative percentage threshold in histogram of found lag times. Using the
            histogram, lagsearch time windows are narrowed down by including bins around
            the peak until a certain percentage of the total values is reached.
        peak_max_count_idx:
            Index of histogram division with max counts
        peak_most_prom_idx
            Index of most prominent histogram peak

        Returns
        -------
        lgs_winsize_adj: Adjusted (narrowed down) time window for lag search
        start_idx: Index of histogram division where the adjusted time window start
        end_idx: Index of histogram division where the adjusted time window ends
        """
        start_idx, end_idx = self.include_bins_next_to_peak(peak_max_count_idx=peak_max_count_idx,
                                                            peak_most_prom_idx=peak_most_prom_idx)

        counts_total = np.sum(counts)
        perc = 0
        while perc < perc_threshold:
            start_idx = start_idx - 1 if start_idx > 0 else start_idx
            end_idx = end_idx + 1 if end_idx < len(counts) else end_idx
            c = counts[start_idx:end_idx]
            perc = np.sum(c) / counts_total
            # print(f"Expanding lag window: {perc}  from record: {start_idx}  to record: {end_idx}")
            if (start_idx == 0) and (end_idx == len(counts)):
                break
        lgs_winsize_adj = [divisions[start_idx], divisions[end_idx]]
        lgs_winsize_adj = [int(x) for x in
                           lgs_winsize_adj]  # Convert elements in array to integers, needed for indexing

        return lgs_winsize_adj, start_idx, end_idx

    def include_bins_next_to_peak(self, peak_max_count_idx, peak_most_prom_idx):
        """Include histogram bins next to the bin for which max was found and the
        most prominent bin.

        Since multiple max count peaks can be detected in the histogram, all found
        peaks are considered and all bins before and after each detected peak are
        included to calculate the adjusted start and end indices.

        For examples:
            Three peaks were with max count were found in the histogram. The peaks
            were found in bins 5, 9 and 14:
                peak_max_count_index = [5,9,14]
            The most prominent peak was detected in bin 2:
                peak_most_prom_idx = 2
            Then the bins before the max count peaks are included:
                start_idx = [4,5,8,9,13,14]
            Then the bins after the max count peaks are included:
                end_idx = [4,5,6,8,9,10,13,14,15]
            Then the max count peaks are combined with the most prominent peak,
            using np.unique() in case of overlapping bins:
                start_end_idx = [2,4,5,6,8,9,10,13,14,15]
            The start_idx is the min of this collection:
                start_idx = 2
            The end_idx is the max of this collection:
                end_idx = 15
            The adjusted time window for lag search starts with the starting time
            of bin 2 and ends with the end time with bin 15 (starting time is added
            in next steps).
        """
        start_idx = np.subtract(peak_max_count_idx, 1)  # Include bins before each peak
        start_idx[start_idx < 0] = 0  # Negative index not possible
        end_idx = np.add(peak_max_count_idx, 1)  # Include bins after each peak
        start_end_idx = np.unique(np.concatenate([start_idx, end_idx, [peak_most_prom_idx]]))  # Combine peaks
        start_idx = np.min(start_end_idx)
        end_idx = np.max(start_end_idx[-1])
        return start_idx, end_idx

    def plot_results_hist(self, hist_bins):
        """
        Plot histogram of found lag times

        Parameters
        ----------
        hist_bins: numpy array
            Histogram bins

        Returns
        -------
        None
        """

        gs, fig, ax = plot.setup_fig_ax()

        # Counts
        # bar_positions = plot_df_agg.index + 0.5  # Shift by half position
        bar_width = (hist_bins[1] - hist_bins[0]) * 0.9  # Calculate bar width
        args = dict(width=bar_width, align='edge')
        ax.bar(x=hist_bins[0:-1], height=self.counts, label='counts', zorder=90, color='#78909c', **args)
        ax.set_xlim(hist_bins[0], hist_bins[-1])

        # Text counts
        for i, v in enumerate(self.counts):
            if v > 0:
                ax.text(hist_bins[0:-1][i] + (bar_width / 2), v, str(v), zorder=99, size=6)

        ax.bar(x=hist_bins[self.peak_max_count_idx], height=self.counts[self.peak_max_count_idx],
               label='most counts', zorder=98, edgecolor='#ef5350', linewidth=4,
               color='None', alpha=0.9, linestyle='-', **args)

        if self.peak_most_prom_idx:
            ax.bar(x=hist_bins[self.peak_most_prom_idx], height=self.counts[self.peak_most_prom_idx],
                   label='most prominent counts peak', zorder=99, edgecolor='#FFA726', linewidth=2,
                   color='None', alpha=0.9, linestyle='--', **args)

        ax.axvline(x=hist_bins[self.start_idx], ls='--', c='#42A5F5',
                   label='lag search window start for next iteration')
        ax.axvline(x=hist_bins[self.end_idx], ls='--', c='#AB47BC',
                   label='lag search window end for next iteration')

        txt_info = \
            f"Histogram of found lag times in iteration {self.iteration}\n" \
            f"Number of found lag times: {self.numvals_series}"

        if self.remove_fringe_bins:
            txt_info += "\nFringe bins removed: yes"
        else:
            txt_info += "\nFringe bins removed: no"

        ax.text(0.02, 0.98, txt_info,
                horizontalalignment='left', verticalalignment='top',
                transform=ax.transAxes, size=14, color='black',
                backgroundcolor='none', zorder=100)

        ax.legend(loc='upper right', prop={'size': 10}).set_zorder(100)

        plot.default_format(ax=ax, label_color='black', fontsize=12,
                            txt_xlabel='lag [records]', txt_ylabel='counts', txt_ylabel_units='[#]')

        if self.outdir:

            if self.remove_fringe_bins:
                outfile = f'{self.iteration}_HISTOGRAM_segment_lag_times_iteration-{self.iteration}.png'
            else:
                outfile = f'HISTOGRAM_segment_lag_times_FINAL.png'

            outpath = self.outdir / outfile
            fig.savefig(f"{outpath}", format='png', bbox_inches='tight', facecolor='w',
                        transparent=True, dpi=150)
            plt.close(fig)

        return None



"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020-2025 Lukas Hörtnagl

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""

import argparse
from pathlib import Path


def validate_args(args):
    """Check validity of optional args"""
    if args.limitnumfiles < 0:
        raise argparse.ArgumentTypeError("LIMITNUMFILES must be 0 or a positive integer.")
    if args.lsnumiter < 1:
        raise argparse.ArgumentTypeError("LSNUMITER must be > 1.")
    if (args.lspercthres < 0.1) | (args.lspercthres > 1):
        raise argparse.ArgumentTypeError("LSPERCTHRES must be between 0.1 and 1.")
    if args.lssegmentduration > args.fileduration:  # todo
        raise argparse.ArgumentTypeError("LSSEGMENTDURATION must be shorter or equal to FILEDURATION.")
    if not args.lssegmentduration:
        # If not specified, then lag times are determined using all of the file data
        args.lssegmentduration = args.fileduration
    if args.lsnumiter <= 0:
        raise argparse.ArgumentTypeError("LSNUMITER must be a positive integer.")
    args.lsremovefringebins = True if args.lsremovefringebins == 1 else False  # Translate settings to bool
    args.delprevresults = True if args.delprevresults == 1 else False  # Translate settings to bool
    return args


def get_args():
    """Get args from CLI input"""
    parser = argparse.ArgumentParser(description="dyco - dynamic lag compensation",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # Positional args
    parser.add_argument('var_reference', type=str,
                        help="Column name of the unlagged reference variable in the data files (one-row header). "
                             "Lags are determined in relation to this signal.")
    parser.add_argument('var_lagged', type=str,
                        help="Column name of the lagged variable in the data files (one-row header). "
                             "The time lag of this signal is determined in relation to the reference "
                             "signal var_reference.")
    parser.add_argument('var_target', nargs='+',
                        help="Column name(s) of the target variable(s). "
                             "Column names of the variables the lag that was found between "
                             "var_reference and var_lagged should be applied to. "
                             "Example: var1 var2 var3")

    # Optional args
    parser.add_argument('-i', '--indir', type=Path,
                        help="Path to the source folder that contains the data files, e.g. 'C:/dyco/input'")
    parser.add_argument('-o', '--outdir', type=Path,
                        help="Path to output folder, e.g. C:/bico/output")
    parser.add_argument('-fnd', '--filenamedateformat', type=str, default='%Y%m%d%H%M%S',
                        help="Filename date format as datetime format strings. Is used to parse the date and "
                             "time info from the filename of found files. The filename(s) of the files found in "
                             "INDIR must contain datetime information. Example for data files named like "
                             "20161015123000.csv: %%Y%%m%%d%%H%%M%%S")
    parser.add_argument('-fnp', '--filenamepattern', type=str, default='*.csv',
                        help="Filename pattern for raw data file search, e.g. *.csv")
    parser.add_argument('-flim', '--limitnumfiles', type=int, default=0,
                        help="Defines how many of the found files should be used. Must be 0 or a positive "
                             "integer. If set to 0, all found files will be used. ")
    parser.add_argument('-fgr', '--filegenres', type=str, default='30T',
                        help="File generation resolution. Example for data files that were generated "
                             "every 30 minutes: 30min")
    parser.add_argument('-fdur', '--fileduration', type=str, default='30T',
                        help="Duration of one data file. Example for data files containing 30 minutes "
                             "of data: 30T")
    parser.add_argument('-dtf', '--datatimestampformat', type=str, default='%Y-%m-%d %H:%M:%S.%f',
                        help="Timestamp format for each row record in the data files. Example for "
                             "high-resolution timestamps like 2016-10-24 10:00:00.024999: "
                             "%%Y-%%m-%%d %%H:%%M:%%S.%%f")
    parser.add_argument('-dres', '--datanominaltimeres', type=float, default=0.05,
                        help="Nominal (expected) time resolution of data records in the files, given as "
                             "one record every x seconds. Example for files recorded at 20Hz: 0.05")
    parser.add_argument('-lss', '--lssegmentduration', type=str, default='30T',
                        help="Segment duration for lag determination. Can be the same as or shorter "
                             "than FILEDURATION.")
    parser.add_argument('-lsw', '--lswinsize', type=int, default=1000,
                        help="Initial size of the time window in which the lag is searched given as "
                             "number of records.")
    parser.add_argument('-lsi', '--lsnumiter', type=int, default=3,
                        help="Number of lag search iterations in Phase 1 and Phase 2. Must be larger than 0.")
    parser.add_argument('-lsf', '--lsremovefringebins', type=int, choices=[0, 1], default=1,
                        help="Remove fringe bins in histogram of found lag times. "
                             "Set to 1 if fringe bins should be removed.")
    parser.add_argument('-lsp', '--lspercthres', type=float, default=0.9,
                        help="Cumulative percentage threshold in histogram of found lag times.")
    parser.add_argument('-lt', '--targetlag', type=int, default=0,
                        help="The target lag given in records to which lag times of all variables "
                             "in var_target are normalized.")
    parser.add_argument('-del', '--delprevresults', type=int, choices=[0, 1], default=0,
                        help="If set to 1, delete all previous results in INDIR. "
                             "If set to 0, search for previously calculated results in "
                             "INDIR and continue.")

    args = parser.parse_args()
    return args



"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020-2025 Lukas Hörtnagl

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""

import datetime as dt
import os
from pathlib import Path

import matplotlib.dates as mdates
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import pandas as pd

from dyco import analyze, files, loop, setup


class SummaryPlots():
    """

    Make plots summarizing results from Phases 1-3

    """

    def __init__(self, instance_phase_1, instance_phase_2, instance_phase_3):
        self.instance_phase_1 = instance_phase_1
        self.instance_phase_2 = instance_phase_2
        self.instance_phase_3 = instance_phase_3

        self.outdir_summary = self._make_outdir()

        self.logger = setup_dyco.create_logger(logfile_path=instance_phase_3.logfile_path, name=__name__)

        self.run()

    def run(self):
        """Start creation of summary plots."""
        self.logger.info("CREATING SUMMARY ...")
        collection_df = pd.DataFrame()

        # Plot and collect time lags and covariances from Phase 1 and Phase 2
        self.logger.info("Plotting found time lags, Phases 1-3")
        phase_instances = [self.instance_phase_1, self.instance_phase_2]
        for phase_instance in phase_instances:
            self.summary_plot_segment_lagtimes(phase=phase_instance.phase,
                                               phase_files=phase_instance.phase_files,
                                               outdirs=phase_instance.outdirs,
                                               last_iteration=phase_instance.lag_n_iter,
                                               target_lag=phase_instance.target_lag)
            collection_df = self._collect_peak_covariances(phase=phase_instance.phase,
                                                           phase_files=phase_instance.phase_files,
                                                           outdirs=phase_instance.outdirs,
                                                           last_iteration=phase_instance.lag_n_iter,
                                                           collection_df=collection_df)

        # Phase 3
        phase_instances = [self.instance_phase_3]
        for phase_instance in phase_instances:
            self.summary_plot_instantaneous_lagtimes(phase=phase_instance.phase,
                                                     phase_files=phase_instance.phase_files,
                                                     outdirs=phase_instance.outdirs)
            collection_df, shift_colname_out = self._collect_instantaneous_lags(phase=phase_instance.phase,
                                                                                phase_files=phase_instance.phase_files,
                                                                                outdirs=phase_instance.outdirs,
                                                                                collection_df=collection_df)
            collection_df = self._collect_instantaneous_covariances(phase=phase_instance.phase,
                                                                    phase_files=phase_instance.phase_files,
                                                                    outdirs=phase_instance.outdirs,
                                                                    collection_df=collection_df,
                                                                    shift_colname_out=shift_colname_out)

        self.summary_plot_covariances(collection_df=collection_df,
                                      outdir=self.outdir_summary,
                                      last_iter_phase_1=self.instance_phase_1.lag_n_iter,
                                      last_iter_phase_2=self.instance_phase_2.lag_n_iter)

        self.logger.info("CREATING SUMMARY ... Done.")
        self.logger.info("Finished DYCO processing.")

    def _add_lines_cov(self, ax, df, last_iter_phase_1, last_iter_phase_2):
        props = {'y': df[f'PHASE-1_ITER-{last_iter_phase_1}_PEAK-COVABSMAX_COV'],
                 'label': f'Phase 1',
                 'zorder': 100, 'alpha': 1, 'fmt': 's', 'ms': 6, 'lw': 1, 'ls': 'dotted',
                 'markeredgecolor': '#80DEEA', 'color': '#80DEEA', 'markerfacecolor': 'white'}
        ax.plot_date(df.index, **props)

        props = {'y': df[f'PHASE-2_ITER-{last_iter_phase_1}_PEAK-COVABSMAX_COV'],
                 'label': f'Phase 2',
                 'zorder': 100, 'alpha': 1, 'fmt': 's', 'ms': 6, 'lw': 1, 'ls': '--',
                 'markeredgecolor': '#F48FB1', 'color': '#F48FB1', 'markerfacecolor': 'white'}
        ax.plot_date(df.index, **props)

        props = {'y': df['PHASE-3_INSTANTANEOUS_LAG_COV'],
                 'label': 'Phase 3',
                 'zorder': 100, 'alpha': 1, 'fmt': 'o', 'ms': 6, 'lw': 2, 'ls': '-',
                 'markeredgecolor': '#8BC34A', 'color': '#8BC34A', 'markerfacecolor': '#8BC34A'}
        ax.plot_date(df.index, **props)

    def _format_plot(self, ax, title, show_legend=True):
        # Automatic tick locations and formats
        locator = mdates.AutoDateLocator(minticks=5, maxticks=20)
        formatter = mdates.ConciseDateFormatter(locator)
        ax.xaxis.set_major_locator(locator)
        ax.xaxis.set_major_formatter(formatter)

        default_format(ax=ax, label_color='black', fontsize=12,
                       txt_xlabel='file date', txt_ylabel='covariance', txt_ylabel_units='')

        font = {'family': 'sans-serif', 'color': 'black', 'weight': 'bold', 'size': 12, 'alpha': 1}
        ax.set_title(title, fontdict=font, loc='left', pad=12)

        ax.axhline(0, color='black', ls='-', lw=1, zorder=1)
        font = {'family': 'sans-serif', 'size': 10}
        if show_legend:
            ax.legend(frameon=True, loc='upper right', prop=font).set_zorder(100)

    def summary_plot_covariances(self, collection_df, outdir, last_iter_phase_1, last_iter_phase_2):
        """Plot covariances found in Phases 1-3."""
        self.logger.info("Plotting covariance evolution, Phases 1-3 ...")
        gs = gridspec.GridSpec(2, 1)  # rows, cols
        gs.update(wspace=0.3, hspace=0.3, left=0.03, right=0.97, top=0.97, bottom=0.03)
        fig = plt.Figure(facecolor='white', figsize=(16, 9))
        ax1 = fig.add_subplot(gs[0, 0])
        ax2 = fig.add_subplot(gs[1, 0])
        props_ax = {'last_iter_phase_1': last_iter_phase_1, 'last_iter_phase_2': last_iter_phase_2}

        # ax1: Covariance evolution
        # =========================
        self._add_lines_cov(ax=ax1, df=collection_df, **props_ax)
        self._format_plot(ax=ax1, title="Half-hourly covariance evolution")

        # ax2: Cumulative
        # ===============
        collection_df_cum = collection_df.cumsum()
        self._add_lines_cov(ax=ax2, df=collection_df_cum, **props_ax)
        self._format_plot(ax=ax2, title="Cumulative covariance evolution", show_legend=False)

        # Save
        outfile = f'summary_covariance_evolution'
        outpath = outdir / outfile
        fig.savefig(f"{outpath}.png", format='png', bbox_inches='tight', facecolor='w', transparent=True, dpi=150)
        plt.close(fig)

    def summary_plot_segment_lagtimes(self, phase: int, phase_files: str, outdirs,
                                      last_iteration: int, target_lag: int):
        """Plot time lags found in last iteration of each Phase incl. aggregated defaults"""
        # Get data
        segment_lagtimes_df = self._get_segment_lagtimes_last_iteration(outdirs=outdirs, phase=phase,
                                                                        phase_files=phase_files,
                                                                        last_iteration=last_iteration)
        lut_df = self._get_lut_last_iteration(outdirs=outdirs, phase=phase, phase_files=phase_files)

        # Make plot
        loop.Loop.plot_segment_lagtimes_ts(segment_lagtimes_df=segment_lagtimes_df,
                                           outdir=self.outdir_summary,
                                           iteration=last_iteration,
                                           show_all=False,
                                           overlay_default=True,
                                           overlay_default_df=lut_df,
                                           overlay_target_val=target_lag,
                                           phase=phase)

    def summary_plot_instantaneous_lagtimes(self, phase, phase_files, outdirs):
        """Plot final reference time lags"""
        df = self._get_instantaneous_time_lags(phase=phase, phase_files=phase_files, outdirs=outdirs)
        analyze.AnalyzeLags.plot_final_instantaneous_lagtimes(outdir=self.outdir_summary, phase=phase, df=df)

    def _collect_instantaneous_covariances(self, outdirs, phase, phase_files, collection_df, shift_colname_out):
        """Get instantaneous covariances at instantaneous time lags from Phase 3"""
        collection_df.index = pd.to_datetime(collection_df.index)
        indir = outdirs[f'{phase}-1_{phase_files}_covariances']
        filelist = os.listdir(str(indir))
        num_files = len(filelist)
        for idx, filename in enumerate(filelist):
            file_date = dt.datetime.strptime(filename, '%Y%m%d%H%M%S_iter1_segment_covariance_iteration-1.csv')

            # Check if datetime of covariance file is contained in collection
            if file_date in collection_df.index:
                # Read covariance file containing all covariances per shift/lag
                filepath = os.path.join(str(indir), filename)
                cov_df = files.read_segment_lagtimes_file(filepath=filepath)

                # From the collection, get the instantaneous shift/lag that was found for this file
                shift_instantaneous = \
                    collection_df.loc[collection_df.index == file_date, shift_colname_out].values[0]

                # From the covariance file, get the covariance for the instantaneous shift/lag
                filter_instantaneous = cov_df['shift'] == shift_instantaneous
                cov_instantaneous = cov_df.loc[filter_instantaneous, 'cov'].values[0]

                # Store instantaneous covariance in collection
                cov_colname_out = f'PHASE-{phase}_INSTANTANEOUS_LAG_COV'
                collection_df.loc[collection_df.index == file_date, cov_colname_out] = cov_instantaneous
        return collection_df

    def _collect_instantaneous_lags(self, phase, phase_files, outdirs, collection_df):
        df = self._get_instantaneous_time_lags(phase=phase, phase_files=phase_files, outdirs=outdirs)
        df.index = pd.to_datetime(df.index)
        shift_colname_out = f'PHASE-{phase}_INSTANTANEOUS_LAG_SHIFT'
        collection_df[shift_colname_out] = df['INSTANTANEOUS_LAG']
        return collection_df, shift_colname_out

    def _get_instantaneous_time_lags(self, phase, outdirs, phase_files):
        """Get instantaneous time lags from Phase 3 LUT file"""
        indir = outdirs[f'{phase}-6_{phase_files}_final_time_lags_lookup_table']
        df = files.read_segment_lagtimes_file(filepath=indir / 'LUT_final_time_lags.csv')
        return df

    def _collect_peak_covariances(self, outdirs, phase, phase_files, last_iteration,
                                  collection_df):

        segment_lagtimes_df = self._get_segment_lagtimes_last_iteration(outdirs=outdirs, phase=phase,
                                                                        phase_files=phase_files,
                                                                        last_iteration=last_iteration)
        _add_df = segment_lagtimes_df[['file_date', 'PEAK-COVABSMAX_COV', 'PEAK-COVABSMAX_SHIFT']]
        _add_df.set_index('file_date', inplace=True)
        cov_colname_out = f'PHASE-{phase}_ITER-{last_iteration}_PEAK-COVABSMAX_COV'
        shift_colname_out = f'PHASE-{phase}_ITER-{last_iteration}_PEAK-COVABSMAX_SHIFT'

        if phase == 1:
            collection_df = _add_df.copy()
            collection_df.rename(columns={'PEAK-COVABSMAX_COV': cov_colname_out}, inplace=True)
            collection_df.rename(columns={'PEAK-COVABSMAX_SHIFT': shift_colname_out}, inplace=True)
        else:
            collection_df[cov_colname_out] = _add_df['PEAK-COVABSMAX_COV']
            collection_df[shift_colname_out] = _add_df['PEAK-COVABSMAX_SHIFT']

        return collection_df

    def _get_segment_lagtimes_last_iteration(self, outdirs, phase, phase_files, last_iteration):
        filepath = outdirs[f'{phase}-3_{phase_files}_time_lags_overview'] \
                   / f'{last_iteration}_segments_found_lag_times_after_iteration-{last_iteration}.csv'
        df = \
            analyze.AnalyzeLags.filter_dataframe(filter_col='iteration',
                                                 filter_equal_to=last_iteration,
                                                 df=files.read_segment_lagtimes_file(
                                                     filepath=filepath))
        return df

    def _get_lut_last_iteration(self, outdirs, phase, phase_files):
        filepath_lut = outdirs[f'{phase}-6_{phase_files}_normalization_lookup_table'] / 'LUT_default_agg_time_lags.csv'
        lut_df = files.read_segment_lagtimes_file(filepath=filepath_lut)
        return lut_df

    def _make_outdir(self):
        outdir = Path(self.instance_phase_1.outdir) / 'SUMMARY'
        if not Path.is_dir(outdir):
            print(f"Creating folder {outdir} ...")
            os.makedirs(outdir)
        return outdir


def default_format(ax, fontsize=12, label_color='black',
                   txt_xlabel='', txt_ylabel='', txt_ylabel_units='',
                   width=1, length=5, direction='in', colors='black', facecolor='white'):
    """Apply default format to plot."""
    ax.set_facecolor(facecolor)
    ax.tick_params(axis='x', width=width, length=length, direction=direction, colors=colors, labelsize=fontsize,
                   top=True)
    ax.tick_params(axis='y', width=width, length=length, direction=direction, colors=colors, labelsize=fontsize,
                   right=True)
    format_spines(ax=ax, color=colors, lw=1)
    if txt_xlabel:
        ax.set_xlabel(txt_xlabel, color=label_color, fontsize=fontsize, fontweight='bold')
    if txt_ylabel and txt_ylabel_units:
        ax.set_ylabel(f'{txt_ylabel}  {txt_ylabel_units}', color=label_color, fontsize=fontsize, fontweight='bold')
    if txt_ylabel and not txt_ylabel_units:
        ax.set_ylabel(f'{txt_ylabel}', color=label_color, fontsize=fontsize, fontweight='bold')


def format_spines(ax, color, lw):
    """Set color and linewidth of axis spines"""
    spines = ['top', 'bottom', 'left', 'right']
    for spine in spines:
        ax.spines[spine].set_color(color)
        ax.spines[spine].set_linewidth(lw)


def setup_fig_ax():
    """Setup grid with one figure and one axis."""
    gs = gridspec.GridSpec(1, 1)  # rows, cols
    gs.update(wspace=0.3, hspace=0.3, left=0.03, right=0.97, top=0.97, bottom=0.03)
    fig = plt.Figure(facecolor='white', figsize=(16, 9))
    ax = fig.add_subplot(gs[0, 0])
    return gs, fig, ax



"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020-2025 Lukas Hörtnagl

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""

import time
from pathlib import Path

import numpy as np
import pandas as pd

from dyco import files


class RemoveLags:
    """
    Remove time lags: use look-up table to normalize time lags across files
    """

    def __init__(self, files_overview_df, data_timestamp_format, outdirs, var_target, lag_n_iter, logger, lut):
        self.files_overview_df = files_overview_df
        self.dat_recs_timestamp_format = data_timestamp_format
        self.outdirs = outdirs
        self.var_target = var_target
        self.lgs_num_iter = lag_n_iter
        self.logger = logger
        self.lut_df = lut

        self.lut_col = 'correction'

        self.run()

    def run(self):
        """Run lag removal"""

        # Loop files
        num_files = self.files_overview_df['file_available'].sum()
        times_needed = []
        files_counter = 0

        for file_idx, file_info_row in self.files_overview_df.iterrows():
            start = time.time()
            txt_info = ""

            # Check file availability
            if file_info_row['file_available'] == 0:
                continue

            this_date = file_info_row['start'].date()
            this_date = pd.to_datetime(this_date)
            shift_correction = self.lut_df.loc[this_date][self.lut_col]

            # Read and prepare data file
            data_df = files.read_raw_data(filepath=file_info_row['filepath'],
                                          data_timestamp_format=self.dat_recs_timestamp_format)  # nrows for testing

            shift_correction = int(shift_correction)
            data_df = self.shift_var_target(df=data_df, shift=shift_correction)

            # No timestamp for final output files
            self.save_dyco_files(outdir=self.outdirs[f'8_time_lags_corrected_files'],
                                 original_filename=file_info_row['filename'],
                                 df=data_df,
                                 export_timestamp=False)

            time_needed = time.time() - start
            times_needed.append(time_needed)
            files_counter += 1
            times_needed_mean = np.mean(times_needed)
            remaining_files = num_files - files_counter
            remaining_sec = times_needed_mean * remaining_files
            progress = (files_counter / num_files) * 100
            txt_info += f"File #{files_counter}: {file_info_row['filename']}" \
                        f"    shift correction: {shift_correction}    remaining time: {remaining_sec:.0f}s" \
                        f"    remaining files: {int(remaining_files)}    progress: {progress:.2f}%"
            self.logger.info(txt_info)

    def save_dyco_files(self, df, outdir, original_filename, export_timestamp):
        """
        Save lag-removed raw data files as new files with the suffix _DYCO

        Parameters
        ----------
        df: pandas DataFrame
            Data of the raw data file where the lag was removed.
        outdir: Path
            Folder where lag-removed raw data files will be stored.
        original_filename: str
            The original filename of the raw data file.
        export_timestamp: bool
            Defines if the timestamp index is saved to the files.

        Returns
        -------
        None
        """
        df.fillna(-9999, inplace=True)
        outpath = outdir / f"{Path(original_filename).stem}_DYCO.csv"
        df.to_csv(outpath, index=export_timestamp)

    def shift_var_target(self, df, shift):
        """
        Shift data of target columns by found lag

        Parameters
        ----------
        df: pandas DataFrame
        shift: int
            Amount by which target columns are shifted, given as number of records.

        Returns
        -------
        pandas DataFrame of shifted raw data
        """
        for col in self.var_target:
            outcol = f"{col}_DYCO"
            df[outcol] = df[col].shift(shift)  # Shift col by found lag
            # df.drop([col], axis=1, inplace=True)  # Remove col that was not shifted
        return df

    def read_lut_time_lags(self):
        """Read csv file that contains the look-up table for lag correction"""
        filepath = self.outdirs[
                       f'6_normalization_lookup_table'] / f'LUT_default_agg_time_lags.csv'
        # parse = lambda x: dt.datetime.strptime(x, '%Y-%m-%d')  # now deprecated
        df = pd.read_csv(filepath,
                         skiprows=None,
                         header=0,
                         # names=header_cols_list,
                         # na_values=-9999,
                         encoding='utf-8',
                         delimiter=',',
                         # mangle_dupe_cols=True,
                         # keep_date_col=False,
                         parse_dates=True,
                         # date_parser=parse,  # now deprecated
                         date_format='%Y-%m-%d',
                         index_col=0,
                         dtype=None,
                         engine='c')
        return df
